// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#ifndef PLCEDITOR_IS_INCLUDED
#define PLCEDITOR_IS_INCLUDED
/**
 * @file plceditor.h
 * @brief An editor for a Piecewise Linear Complex to be used by tetgenbuilder
 *
 * @author Johan Hake
 * @date 2007-07-17
 */

#include <vector>
#include <map>
#include <stdexcept>
#include <iostream>

#include "plc.h"

namespace tritetmesh{

  class PLCEditor
  {

  public:

    /// Constructor
    ~PLCEditor();
  
    /// Destructor
    PLCEditor();

    /// Clear the registered PLC data
    void clear();

    /// Print all vertices
    void print_vertices();
    
    /**
     * @brief Register vertices by their coordinates
     *
     * Add vertices to the PLC by giving the coordinates in three different 
     * array(vector) parameters.
     *
     * The coordinate lists have to be the same size.
     *
     * @param x_coord A sequence with the x coordinates
     * @param y_coord A sequence with the y coordinates
     * @param z_coord A sequence with the z coordinates
     */
    void register_vertices(const std::vector<double> &x_coord, 
			   const std::vector<double> &y_coord, 
			   const std::vector<double> &z_coord);
    
    /**
     * @brief Add a marker to the supplied vector of vertices
     *
     * @param marker The marker (integer)
     * @param vertices A sequence with the vertices 
     * 
     */
    void add_vertex_marker(unsigned int marker, const std::vector<unsigned int> &vertices);

    /**
     * @brief Add a new facet to the PLC.
     *
     * The facet is added by registering a polygon. This is provided as a sequence of vertex numbers.
     * Additional polygons can be added to this facet by the function add_polygon_to_last_facet()
     *
     * Optionally a facet marker and a maximum area for the sub faces on this facet can be provided.
     *
     * @param polygon A sequence of vertex numbers defining the polygon
     * @param marker An integer number that is inherited by subfaces of this facet
     * @param max_area A positive float that sets an upper bound on the area of each sub face of this facet
     */
    void add_facet(const std::vector<unsigned int> &polygon, unsigned int marker = 0, double max_area = -1.0);
  
    /**
     * @brief Add a polygon to the last registered facet.
     *
     * The polygon is provided as a sequence of vertex numbers
     *
     * @param polygon A sequence of vertex numbers defining the polygon
     */
    void add_polygon_to_last_facet(const std::vector<unsigned int> &polygon);
  
    /**
     * @brief Add a polygon and a hole to the last registered facet
     *
     * The polygon is provided as a sequence of vertex numbers and the hole is provided as a point situated on the facet.
     *
     * @param polygon A sequence of vertex numbers defining the polygon
     * @param point   A sequence of length 3, containing the coordinates of the point
     */
    void add_polygon_to_last_facet(const std::vector<unsigned int> &polygon, const std::vector<double> &point);

    /**
     * @brief Add a polygon to a facet
     *
     * The polygon is provided as a sequence of vertex numbers.
     * If as_a_hole is true then the polygon is added as a hole, 
     * using the mid point values of the vertices
     *
     * @param polygon A sequence of vertex numbers defining the polygon
     * @param facet_ind  The facet index
     * @param as_a_hole True if the polygon should be interpreted as a facet hole
     */
    void add_polygon_to_facet(const std::vector<unsigned int> &polygon, 
			      unsigned int facet_ind, 
			      bool as_a_hole = false);
    /**
     * @brief Add a hole to the last registered facet.
     *
     * The hole is provided as a point situated on the facet.
     * 
     * @param point   A sequence of length 3, containing the coordinates of the point
     */
    void add_facet_hole(const std::vector<double> &point);
  
    /**
     * @brief Add a hole to the last registered facet.
     *
     * The hole is provided as a point situated on the facet.
     * 
     * @param point  A sequence of length 3, containing the coordinates of the point
     * @param facet_ind The facet index
     */
    void add_facet_hole(const std::vector<double> &point, unsigned int facet_ind);
  
    /**
     * @brief Add a marker to the last registered facet
     *
     * @param marker  An integer number that is inherited by subfaces of this facet
     */
    void add_facet_marker(unsigned int marker);

    /**
     * @brief Add a marker to a given facet.
     *
     * @param marker An integer number that is inherited by subfaces of this facet
     * @param facet_ind  An integer number defining the facet.
     */
    void add_facet_marker(unsigned int marker, unsigned int facet_ind);

    /**
     * @brief Add an area constraint for the subfaces of the last registered facet.
     *
     * @param max_area A positive float that sets an upper bound on the area of each sub face of this facet
     */
    void add_facet_area_constraints(double max_area);

    /**
     * @brief Add an area constraint for the subfaces of a given facet
     *
     * @param max_area A positive float that sets an upper bound on the area of each sub face of this facet
     * @param facet_ind  An integer number defining the facet.
     */
    void add_facet_area_constraints(double max_area, unsigned int facet_ind);

    /**
     * @brief Add a hole to the PLC.
     *
     * The hole is provided as a point situated inside the PLC. 
     * Make sure that the point is sourrounded by facets. The mesh can otherwise dissapear!
     * 
     * @param point   A sequence of length 3, containing the coordinates of the point
     */
    void add_hole(const std::vector<double> &point);

    /**
     * @brief Add a region to the PLC
     *
     * The region is provided as a point situated inside the PLC together with an attribute.
     * All tetrahedrons situated inside the region defined by any sourrounding facets
     * will inherit this attribute.
     *
     * Optionally a maximal volume can be provided limiting the volume of the tetrahedrons 
     * inside the region.
     * 
     * @param point   A sequence of length 3, containing the coordinates of the point
     * @param attribute A float defining the attribute of the tetrahedrons inside the region
     * @param max_volume A float definging the maximal volume of the tetrahedrons inside the region
     */
    void add_region(const std::vector<double> &point, double attribute, double max_volume = -1.0);

    /**
     * @brief Build the mesh
     *
     * The mesh is builded using iether the information in any loaded plc or by the registered
     * plc information.
     * 
     * The mesh is then saved in a dolfin format using the provided file name as base name.
     */
    void build(std::string file_name, bool save_tetgen = false, bool save_off = false);
  
    /**
     * @brief Build the mesh
     *
     * The mesh is builded using either the information in any loaded plc or by the registered
     * plc information.
     */
    void build();
    
    /// Save mesh to file in dolfin format
    void save_mesh(std::string file_name, bool save_tetgen = false, bool save_off = false);
    
    /// Use tetgen to detect self-intersections of facets of the PLC.
    void test_plc();
  
    /// Print tetgen help
    inline void print_tetgen_help(){tetmesh.print_tetgen_help();}

    /// Print the number of registered facets
    inline unsigned int num_facets(){return facet_list.size();}

    /// Wrapper functions for the plc
    
    /// Load plc from file. Accepted file formats are: poly, smesh, off, ply, stl, mesh
    inline void load(char* file_name){clear();plc.load(file_name);plc_builded = true;}

    /// Write plc to file in tetgen format
    inline void save_plc(char* file_name){build_plc();plc.save_tetgen(file_name);plc.clear();plc_builded=false;}

    /// Print parameters for building mesh
    inline void print_parameters(){plc.print_tetgen_parameters();}

    /// Set quality of generated mesh.
    inline void set_radius_edge_ratio(double ratio){plc.set_radius_edge_ratio(ratio);}

    /// Set uniform volume constraint.
    inline void set_uniform_volume_constraint(double constraint)
    {plc.set_uniform_volume_constraint(constraint);}

    /// Set verbosity
    inline void set_verbosity(int verbosity){plc.set_verbosity(verbosity);}

    /// Supress splitting of facets
    inline void set_splitting_of_facets(bool split){plc.set_splitting_of_facets(split);}

  private:
    void build_plc();
    
    bool plc_builded;
    bool tetmesh_builded;

    std::vector<double>  x_coord;
    std::vector<double>  y_coord;
    std::vector<double>  z_coord;
    std::map<unsigned int,unsigned int>    vertex_markers;
    
    std::vector< std::vector<std::vector<unsigned int> > > facet_list;
    std::map<unsigned int,std::vector<double> > facet_holes;
    std::map<unsigned int,unsigned int>    facet_markers;
    std::map<unsigned int,double> facet_area_constraints;
    
    std::vector<double> hole_list;
    std::vector<double> region_list;
    
    tritetmesh::PLC     plc;
    tritetmesh::TetMesh tetmesh;
  };
}

#endif
