// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#ifndef PSLGEDITOR_IS_INCLUDED
#define PSLGEDITOR_IS_INCLUDED
/**
 * @file pslgeditor.h
 * @brief An editor for a Piecewise Linear Complex to be used by tetgenbuilder
 *
 * @author Johan Hake
 * @date 2007-07-17
 */

#include <vector>
#include <set>
#include <map>
#include <stdexcept>
#include <iostream>

extern "C" {
#include <triangle_mesh.h>
}


#include "pslg.h"
#include "areafunction.h"

namespace tritetmesh{

  class PSLGEditor
  {

  public:

    /// Constructor
    ~PSLGEditor();
  
    /// Destructor
    PSLGEditor();

    /// Clear the registered PSLG data
    void clear();

    /// Print all vertices
    void print_vertices();
  
    /**
     * @brief Register vertices by their coordinates
     *
     * Add vertices to the PSLG by giving the coordinates in two different 
     * array(vector) parameters.
     *
     * The coordinates have to be the same size.
     *
     * Optionally several sequences (max 3) can be provided, with attributes that are
     * assigned each vertex. These values are then interpolated in the new mesh.
     *
     * @param x_coord A sequence with the x coordinates
     * @param y_coord A sequence with the y coordinates
     * @param attributesX Optional sequences of attributes that are assigned each vertex
     */
    void register_vertices(const std::vector<double> &x_coord, 
			   const std::vector<double> &y_coord);
    
    /**
     * @brief Register vertices by their coordinates
     *
     * Add vertices to the PSLG by giving the coordinates in two different 
     * array(vector) parameters.
     *
     * The coordinates have to be the same size.
     *
     * Optionally several sequences (max 3) can be provided, with attributes that are
     * assigned each vertex. These values are then interpolated in the new mesh.
     *
     * @param x_coord A sequence with the x coordinates
     * @param y_coord A sequence with the y coordinates
     * @param attributesX Optional sequences of attributes that are assigned each vertex
     */
    void register_vertices(const std::vector<double> &x_coord, 
			   const std::vector<double> &y_coord,
			   const std::vector<double> &attributes);
    
    /**
     * @brief Register vertices by their coordinates
     *
     * Add vertices to the PSLG by giving the coordinates in two different 
     * array(vector) parameters.
     *
     * The coordinates have to be the same size.
     *
     * Optionally several sequences (max 3) can be provided, with attributes that are
     * assigned each vertex. These values are then interpolated in the new mesh.
     *
     * @param x_coord A sequence with the x coordinates
     * @param y_coord A sequence with the y coordinates
     * @param attributesX Optional sequences of attributes that are assigned each vertex
     */
    void register_vertices(const std::vector<double> &x_coord, 
			   const std::vector<double> &y_coord,
			   const std::vector<double> &attributes1,
			   const std::vector<double> &attributes2);

    /**
     * @brief Register vertices by their coordinates
     *
     * Add vertices to the PSLG by giving the coordinates in two different 
     * array(vector) parameters.
     *
     * The coordinates have to be the same size.
     *
     * Optionally several sequences (max 3) can be provided, with attributes that are
     * assigned each vertex. These values are then interpolated in the new mesh.
     *
     * @param x_coord A sequence with the x coordinates
     * @param y_coord A sequence with the y coordinates
     * @param attributesX Optional sequences of attributes that are assigned each vertex
     */
    void register_vertices(const std::vector<double> &x_coord, 
			   const std::vector<double> &y_coord,
			   const std::vector<double> &attributes1, 
			   const std::vector<double> &attributes2,
			   const std::vector<double> &attributes3);
  
    /**
     * @brief Add a marker to the supplied vector of vertices
     *
     * @param marker The attribute (integer)
     * @param vertices A sequence with the vertices 
     * 
     */
    void add_vertex_marker(unsigned int marker, const std::vector<unsigned int> &vertices);
    
    /**
     * @brief Add attributes to the correspondning vertices
     *
     * The sequence provided should be the same length as the number of registered vertices.
     * Any number of vertex attributes can be provided for each vertex. 
     * Call this function several times to facilitate this.
     *
     * @param attributes The attributes given as a sequence (float)
     *
     */
    void add_vertex_attributes(const std::vector<double> &attributes);

    /**
     * @brief Add a polygon to the PSLG.
     *
     * The polygon is given by a sequence of vetex number. 
     * The polygon should be closed or closeable, i.e., not a straight line.
     *
     * Optionally a marker and a maximum area for the triangles in the polygon
     * can be provided.
     * This option is only available for simple polygons, where the mean of 
     * the coordinates is within the border of the polygon, and no other polygons
     * is included inside it.
     * If this is not the case, set the region manually using add_region().
     *
     * @param polygon A sequence of vertex numbers defining the closed polygon
     * @param attribute A float that is inherited by triangles in this polygon
     * @param max_area A positive float that sets an upper bound on the area of each triangle in the polygon
     */
    void add_polygon(const std::vector<unsigned int> &polygon, double attribute = 0, double max_area = -1.0);

    /**
     * @brief Add a polygon to the PSLG and make a hole out of it.
     *
     * The polygon is given by a sequence of vetex number. 
     * The polygon should be closed or closeable, i.e., not a straight line.
     *
     * The polygon need to be simpel, where the mean of the coordinates is within
     * the border of the polygon. 
     * If a more complex polygonal hole is neaded, set the hole manually
     * using add_hole().
     *
     * @param polygon A sequence of vertex numbers defining the closed polygon
     */
    void add_polygon_hole(const std::vector<unsigned int> &polygon);
  
    /**
     * @brief Add a seqence of segments
     *
     * The segments is provided by a sequence of vertex numbers
     * Optionally a marker can be provided that is inherited by the any sub-segments 
     *
     * If the sequence include segments that is allready registered, e.g., by another
     * function, they are not added, but if a marker is provided this is registered.
     *
     * @param sequence A sequence of vertex numbers defining the segments
     * @param marker An integer defining the marker that will be inherited by any sub-segments
     */
    void add_segments(const std::vector<unsigned int> &segments, unsigned int marker = 0);
  
    /**
     * @brief Add a hole to the PSLG.
     *
     * The hole is provided as a point situated inside a closed path of segments, previuosly
     * provded either by add_segments() or add_polygon(). 
     * Make sure that the point is sourrounded by a closed path of segments.
     * The mesh can otherwise dissapear!
     * 
     * @param point A sequence of length 2, containing the coordinates of the point
     */
    void add_hole(const std::vector<double> &point);

    /**
     * @brief Add a region to the PSLG
     *
     * The region is provided as a point situated inside a closed path of segents, previuosly
     * provded either by add_segments() or add_polygon(), together with an attribute.
     * All triangles produced inside the region will inherit the attribute.
     *
     * Optionally a maximal are can be provided limiting the volume of the tetrahedrons 
     * inside the region.
     * 
     * @param point  A sequence of length 2, containing the coordinates of the point
     * @param attribute A float defining the attribute of the tetrahedrons inside the region
     * @param max_volume A float definging the maximal volume of the tetrahedrons inside the region
     */
    void add_region(const std::vector<double> &point, double attribute, double max_area = -1.0);

    /**
     * @brief Build the PSLG.
     *
     * Use the registered information to build a mesh. 
     * Save the result with the provided file name
     */
    void build(std::string file_name, bool save_triangle = false);
  
    // Wrapper functions for pslg functionalities
    
    /// Print the present triangle parameters
    inline void print_triangle_parameters(){pslg.print_triangle_parameters();}
    
    /// Set a minimum angle of a generated triangle.  (triangle -qX option)
    inline void set_minimum_angle(double angle){pslg.set_minimum_angle(angle);}
    
    /// Set uniform volume constraint  (triangle -aX.X option)
    inline void set_uniform_area_constraint(double constraint){pslg.set_uniform_area_constraint(constraint);}
    
    /// Set a user defined area function.
    inline void add_area_function(tritetmesh::AreaFunction &area_func)
    {
      pslg.add_area_function(area_func);
    }
    
    /// Set verbosity  (triangle -V option)
    inline void set_verbosity(int verbosity){pslg.set_verbosity(verbosity);}

    /// set generate edge indformation  (triangle -e option)
    inline void set_generate_edge_information(bool generate)
    {
      pslg.set_generate_edge_information(generate);
    }

    /// Enclose a convex hull (triangle -c option)
    inline void set_enclose_convex_hull(bool enclose)
    {
      pslg.set_enclose_convex_hull(enclose);
    }

    /// Write plc to file in triangle format
    inline void save_pslg(char* file_name){build_pslg();pslg.save_triangle(file_name);pslg.clear();pslg_builded = false;}

  private:
    void build_pslg();
    
    bool pslg_builded;

    std::vector<double>  x_coord;
    std::vector<double>  y_coord;

    std::vector< std::vector<double> > vertex_attributes;
    std::map<unsigned int,unsigned int>    vertex_markers;

    std::set<std::pair<unsigned int,unsigned int> > segment_list;
    std::map<std::pair<unsigned int,unsigned int>,unsigned int> segment_markers;

    std::vector<double> hole_list;
    std::vector<double> region_list;

    tritetmesh::PSLG    pslg;
    tritetmesh::TriMesh trimesh;
  };
}

#endif
